/*
 * wxcmisc.cc
 * wxWindows miscellaneous command processing
 *
 */

// For compilers that support precompilation, includes "wx.h".
#include "wx_prec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx.h"
#endif

#include "wx_hash.h"
#include "wx_types.h"
#include "wx_timer.h"

#include <iostream.h>
#include <ctype.h>
#include <stdlib.h>

#include "wx_cmds.h"
#include "wxcmdsp.h"

#ifdef USE_CLIPS
#ifndef __WATCOMC__
#include "clips.h"
#endif
#endif

extern "C" int clipsEchoOn;

// Helper functions to associate a callback list with an object, NOT using
// client data as we did originally, but using a hash table for generality
// to non-window objects.
// static wxHashTable wxClipsCallbackHashTable(wxKEY_INTEGER);

void wxClipsAddCallback(WXID id, char *eventType, wxClipsFunction callback)
{
/*
  wxClipsCallbacks *list = (wxClipsCallbacks *)wxClipsCallbackHashTable.Get(id);

  if (!list)
  {
    list = new wxClipsCallbacks;
    wxClipsCallbackHashTable.Put(id, list);
  }
*/
  wxClipsCallbacks *list = (wxClipsCallbacks *)wxGetUserData(id);
  if (!list)
  {
    list = new wxClipsCallbacks;
    wxAssociateUserData(id, (void *)list);
  }

#ifdef USE_CLIPS
  list->Append(eventType, (wxObject *)copystring(callback));
#else
  list->Append(eventType, (wxObject *)callback);
#endif
}

wxClipsFunction wxClipsFindCallback(WXID id, char *eventType)
{
//   wxClipsCallbacks *list = (wxClipsCallbacks *)wxClipsCallbackHashTable.Get(id);
  wxClipsCallbacks *list = (wxClipsCallbacks *)wxGetUserData(id);
  if (list)
  {
    wxNode *node = list->Find(eventType);
    if (node)
      return (wxClipsFunction)node->Data();
  }
  return (wxClipsFunction)NULL;
}

void wxClipsDeleteCallbacks(WXID id)
{
  wxClipsCallbacks *list = (wxClipsCallbacks *)wxGetUserData(id);
  if (list)
    delete list;
}

/*
 * Per-class callback for object deletion: removes the per-object callback list
 * from the hash table.
 */
 
void wxClipsObjectOnDelete(WXID id)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnDelete");
  if (func)
  {
#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld)", func, id);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long)) func) (id);
#endif
  }

  wxClipsDeleteCallbacks(id);
}

/*
 * OnCharHook callback for wxApp class.
 */
#ifdef wx_msw 
Bool wxClipsAppClassOnCharHook(Bool *handled, WXID id, WXID eventId)
{
  *handled = FALSE;
  
  wxClipsFunction func = wxClipsFindCallback(id, "OnCharHook");
  if (func)
  {
#ifdef USE_CLIPS
    char buf[100];
    clipsReturnType = clipsUNKNOWN;
    sprintf(buf, "(return-result (%s %ld %ld))", func, id, eventId);
    wxRouteNoEcho(buf);
#else
    clipsReturnLong = (* (long (*)(long, long)) func) (id, eventId);
    clipsReturnType = clipsLONG;
#endif
    if (clipsReturnType != clipsLONG)
    {
      ExtensionError << "OnCharHook callback should return an integer (1 or 0)!\n";
      return FALSE;
    }
    else
    {
      *handled = TRUE;
      return (Bool)clipsReturnLong;
    }
  }
  return FALSE;
}
#endif

/*
 * Mechanism for getting CLIPS callbacks to return a value
 *
 */

int clipsReturnType = clipsUNKNOWN;
char *clipsReturnString = NULL;
long clipsReturnLong = 0;
double clipsReturnDouble = 0.0;

/*
 * The argument to an externally called function is passed to this
 * function, which stores the appropriate return value and type
 * for the external program to fetch.
 */

#ifdef USE_CLIPS
void clipsReturnResult()
{
  DATA_OBJECT temp;
  RtnUnknown(1, &temp);

  switch (ClipsGetType(temp))
  {
    case INTEGER:
    {
      clipsReturnLong = DOToLong(temp);
      clipsReturnType = clipsLONG;
      break;
    }
    case STRING:
    case SYMBOL:
    {
      if (clipsReturnString)
        delete[] clipsReturnString;

      clipsReturnString = copystring(DOToString(temp));
      clipsReturnType = clipsSTRING;
      break;
    }
    case FLOAT:
    {
      clipsReturnDouble = DOToDouble(temp);
      clipsReturnType = clipsDOUBLE;
      break;
    }
    default:
    {
      clipsReturnType = clipsUNKNOWN;
    }
  }
}

char *clipsGetPlatform()
{
  return (char *)AddSymbol(wxcGetPlatform());
}

char *clipsGetOsVersion()
{
  return (char *)AddSymbol(wxcGetOsVersion());
}

/************************************************************************
 * CLEARING-UP FUNCTIONS
 *
 ************************************************************************/

void clipsCleanWindows()
{
  if (ArgCountCheck("clipsCleanWindows", EXACTLY, 0) == -1)
    return;

  wxcCleanWindows();
}
#endif

/*
 * Timer functions
 */
 
void wxClipsTimerClassNotify(Bool *handled, WXID id)
{
  wxClipsFunction func = wxClipsFindCallback(id, "Notify");
  if (func)
  {
    *handled = TRUE;
        
#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld)", func, id);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long)) func) (id);
#endif
  }
}

#ifdef USE_CLIPS
long clipsTimerCreate()
{
  if (ArgCountCheck("clipsTimerCreate", EXACTLY, 0) == -1)
    return 0;

  return wxcTimerCreate();
}

long clipsTimerDelete()
{
  if (ArgCountCheck("clipsTimerDelete", EXACTLY, 1) == -1)
    return 0;

  return (long)wxcTimerDelete(RtnLong(1));
}

long clipsTimerStart()
{
  if (ArgCountCheck("clipsTimerStart", EXACTLY, 2) == -1)
    return 0;

  return wxcTimerStart(RtnLong(1), RtnLong(2));
}

long clipsTimerStop()
{
  if (ArgCountCheck("clipsTimerStop", EXACTLY, 1) == -1)
    return 0;

  return wxcTimerStop(RtnLong(1));
}

/*
 * Stopwatch functions
 */
 
void clipsStartTimer()
{
  if (ArgCountCheck("clipsStartTimer", EXACTLY, 0) == -1)
    return;

  wxcStartTimer();
}

long clipsGetElapsedTime()
{
  if (ArgCountCheck("clipsGetElapsedTime", AT_LEAST, 0) == -1)
    return 0;

  Bool resetTimer = FALSE;
  if (RtnArgCount() > 0)
    resetTimer = (Bool)RtnLong(0);
  return wxcGetElapsedTime(resetTimer);
}

char *clipsNow()
{
  if (ArgCountCheck("clipsNow", EXACTLY, 0) == -1)
    return (char *)AddSymbol("");

  char *s = wxcNow();
  if (s)
    return (char *)AddSymbol(s);
  else
    return (char *)AddSymbol("");
}

// wxApp functions

#ifdef USE_CLIPS
long clipsAppCreate()
{
  if (ArgCountCheck("clipsAppCreate", EXACTLY, 0) == -1)
    return 0;

  return wxcAppCreate();
}

void clipsAppSetShowFrameOnInit()
{
  if (ArgCountCheck("clipsAppSetShowFrameOnInit", EXACTLY, 2) == -1)
    return;

  wxcAppSetShowFrameOnInit((WXID)RtnLong(1), (Bool)RtnLong(2));
}

long clipsAppGetShowFrameOnInit()
{
  if (ArgCountCheck("clipsAppGetShowFrameOnInit", EXACTLY, 1) == -1)
    return 0;

  return (long)wxcAppGetShowFrameOnInit((WXID)RtnLong(1));
}
#endif

/*
 * Mouse event functions
 */
 
long clipsMouseEventIsButton()
{
  if (ArgCountCheck("clipsMouseEventIsButton", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventIsButton(id);
}

long clipsMouseEventButtonDown()
{
  if (ArgCountCheck("clipsMouseEventButtonDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventButtonDown(id);
}

long clipsMouseEventControlDown()
{
  if (ArgCountCheck("clipsMouseEventControlDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventControlDown(id);
}

long clipsMouseEventShiftDown()
{
  if (ArgCountCheck("clipsMouseEventShiftDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventShiftDown(id);
}

long clipsMouseEventButton()
{
  if (ArgCountCheck("clipsMouseEventButton", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long butt = RtnLong(2);
  return wxcMouseEventButton(id, (int)butt);
}

long clipsMouseEventLeftDown()
{
  if (ArgCountCheck("clipsMouseEventLeftDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventLeftDown(id);
}

long clipsMouseEventMiddleDown()
{
  if (ArgCountCheck("clipsMouseEventMiddleDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventMiddleDown(id);
}

long clipsMouseEventRightDown()
{
  if (ArgCountCheck("clipsMouseEventRightDown", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventRightDown(id);
}

long clipsMouseEventLeftUp()
{
  if (ArgCountCheck("clipsMouseEventLeftUp", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventLeftUp(id);
}

long clipsMouseEventMiddleUp()
{
  if (ArgCountCheck("clipsMouseEventMiddleUp", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventMiddleUp(id);
}

long clipsMouseEventRightUp()
{
  if (ArgCountCheck("clipsMouseEventRightUp", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventRightUp(id);
}

long clipsMouseEventDragging()
{
  if (ArgCountCheck("clipsMouseEventDragging", EXACTLY, 1) == -1)
    return 0;
    
  long id = RtnLong(1);
  return wxcMouseEventDragging(id);
}

double clipsMouseEventPositionX()
{
  if (ArgCountCheck("clipsMouseEventPositionX", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventPositionX(id);
}

double clipsMouseEventPositionY()
{
  if (ArgCountCheck("clipsMouseEventPositionY", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcMouseEventPositionY(id);
}

/*
 * Key event functions
 */
 
char *clipsKeyEventGetKeyCode()
{
  if (ArgCountCheck("clipsKeyEventGetKeyCode", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

#if wxRELEASE_NUMBER > 3
  char *s = wxExtensionKeyCodeIntToString(wxcKeyEventKeyCode(RtnLong(1)));
#else
  char *s = wxcKeyEventKeyCodeString(RtnLong(1));
#endif
  if (s)
    return (char *)AddSymbol(s);
  else
    return (char *)AddSymbol("");
}

double clipsKeyEventPositionX()
{
  if (ArgCountCheck("clipsKeyEventPositionX", EXACTLY, 1) == -1)
    return 0.0;
  double x = 0.0;
  double y = 0.0;
  wxcKeyEventPosition(RtnLong(1), &x, &y);
  return x;
}

double clipsKeyEventPositionY()
{
  if (ArgCountCheck("clipsKeyEventPositionY", EXACTLY, 1) == -1)
    return 0.0;
  double x = 0.0;
  double y = 0.0;
  wxcKeyEventPosition(RtnLong(1), &x, &y);
  return y;
}

long clipsKeyEventControlDown()
{
  if (ArgCountCheck("clipsKeyEventControlDown", EXACTLY, 1) == -1)
    return 0;
  return (long)wxcKeyEventControlDown(RtnLong(1));
}

long clipsKeyEventShiftDown()
{
  if (ArgCountCheck("clipsKeyEventShiftDown", EXACTLY, 1) == -1)
    return 0;
  return (long)wxcKeyEventShiftDown(RtnLong(1));
}

long clipsKeyEventAltDown()
{
  if (ArgCountCheck("clipsKeyEventAltDown", EXACTLY, 1) == -1)
    return 0;
  return (long)wxcKeyEventAltDown(RtnLong(1));
}

/*
 * Generic event functions
 */

char *clipsEventGetEventType()
{
  if (ArgCountCheck("clipsGetEventType", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  long eventType = wxcEventGetEventType(id);
  char *eventTypeName = NULL;
  switch (eventType)
  {
    case wxEVENT_TYPE_BUTTON_COMMAND: eventTypeName = "wxEVENT_TYPE_BUTTON_COMMAND";
      break;
    case wxEVENT_TYPE_CHECKBOX_COMMAND: eventTypeName = "wxEVENT_TYPE_CHECKBOX_COMMAND";
      break;
    case wxEVENT_TYPE_CHOICE_COMMAND: eventTypeName = "wxEVENT_TYPE_CHOICE_COMMAND";
      break;
    case wxEVENT_TYPE_LISTBOX_COMMAND: eventTypeName = "wxEVENT_TYPE_LISTBOX_COMMAND";
      break;
    case wxEVENT_TYPE_TEXT_COMMAND: eventTypeName = "wxEVENT_TYPE_TEXT_COMMAND";
      break;
    case wxEVENT_TYPE_TEXT_ENTER_COMMAND: eventTypeName = "wxEVENT_TYPE_TEXT_ENTER_COMMAND";
      break;
    case wxEVENT_TYPE_MULTITEXT_COMMAND: eventTypeName = "wxEVENT_TYPE_MULTITEXT_COMMAND";
      break;
    case wxEVENT_TYPE_MENU_COMMAND: eventTypeName = "wxEVENT_TYPE_MENU_COMMAND";
      break;
    case wxEVENT_TYPE_SLIDER_COMMAND: eventTypeName = "wxEVENT_TYPE_SLIDER_COMMAND";
      break;
    case wxEVENT_TYPE_RADIOBOX_COMMAND: eventTypeName = "wxEVENT_TYPE_RADIOBOX_COMMAND";
      break;
    case wxEVENT_TYPE_SET_FOCUS: eventTypeName = "wxEVENT_TYPE_SET_FOCUS";
      break;
    case wxEVENT_TYPE_KILL_FOCUS: eventTypeName = "wxEVENT_TYPE_KILL_FOCUS";
      break;
    case wxEVENT_TYPE_LEFT_DOWN: eventTypeName = "wxEVENT_TYPE_LEFT_DOWN";
      break;
    case wxEVENT_TYPE_LEFT_UP: eventTypeName = "wxEVENT_TYPE_LEFT_UP";
      break;
    case wxEVENT_TYPE_MIDDLE_DOWN: eventTypeName = "wxEVENT_TYPE_MIDDLE_DOWN";
      break;
    case wxEVENT_TYPE_RIGHT_DOWN: eventTypeName = "wxEVENT_TYPE_RIGHT_DOWN";
      break;
    case wxEVENT_TYPE_RIGHT_UP: eventTypeName = "wxEVENT_TYPE_RIGHT_UP";
      break;
    case wxEVENT_TYPE_MOTION: eventTypeName = "wxEVENT_TYPE_MOTION";
      break;
    case wxEVENT_TYPE_ENTER_WINDOW: eventTypeName = "wxEVENT_TYPE_ENTER_WINDOW";
      break;
    case wxEVENT_TYPE_LEAVE_WINDOW: eventTypeName = "wxEVENT_TYPE_LEAVE_WINDOW";
      break;
    case wxEVENT_TYPE_LEFT_DCLICK: eventTypeName = "wxEVENT_TYPE_LEFT_DCLICK";
      break;
    case wxEVENT_TYPE_MIDDLE_DCLICK: eventTypeName = "wxEVENT_TYPE_MIDDLE_DCLICK";
      break;
    case wxEVENT_TYPE_RIGHT_DCLICK: eventTypeName = "wxEVENT_TYPE_RIGHT_DCLICK";
      break;
    default:
      break;
  }
  if (eventTypeName)
    return (char *)AddSymbol(eventTypeName);
  else
    return (char *)AddSymbol("");
}

/*
 * Command event functions
 */

long clipsCommandEventIsSelection()
{
  if (ArgCountCheck("clipsCommandEventIsSelection", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcCommandEventIsSelection(id);
}

long clipsCommandEventGetSelection()
{
  if (ArgCountCheck("clipsCommandEventGetSelection", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcCommandEventGetSelection(id);
}

#endif

/************************************************************************
 * FUNCTIONS TO ACCESS HELP SYSTEM
 *
 ************************************************************************/

long clipsHelpCreate()
{
  if (ArgCountCheck("clipsHelpCreate", AT_LEAST, 0) == -1)
    return 0;
  Bool native = TRUE;
  if (RtnArgCount() > 0)
    native = (Bool)RtnLong(1);
  return (long)wxcHelpCreate(native);
}

long clipsHelpDelete()
{
  if (ArgCountCheck("clipsHelpDelete", EXACTLY, 1) == -1)
    return 0;
  long id = RtnLong(1);
  return wxcHelpDelete(id);
}

long clipsHelpLoadFile()
{
  if (ArgCountCheck("clipsHelpLoadFile", EXACTLY, 2) == -1)
    return 0;
  long id = RtnLong(1);
  char *filename = RtnLexeme(2);
  if (!filename)
    return 0;
  return wxcHelpLoadFile(id, filename);
}

long clipsHelpDisplayContents()
{
  if (ArgCountCheck("clipsHelpDisplayContents", EXACTLY, 1) == -1)
    return 0;
  return wxcHelpDisplayContents(RtnLong(1));
}

long clipsHelpDisplaySection()
{
  if (ArgCountCheck("clipsHelpDisplaySection", EXACTLY, 2) == -1)
    return 0;
  return wxcHelpDisplaySection(RtnLong(1), RtnLong(2));
}

long clipsHelpDisplayBlock()
{
  if (ArgCountCheck("clipsHelpDisplayBlock", EXACTLY, 2) == -1)
    return 0;
  return wxcHelpDisplayBlock(RtnLong(1), RtnLong(2));
}

long clipsHelpKeywordSearch()
{
  if (ArgCountCheck("clipsHelpKeywordSearch", EXACTLY, 2) == -1)
    return 0;
  long id = RtnLong(1);
  char *keyword = RtnLexeme(2);
  if (!keyword)
    return 0;
  return wxcHelpKeywordSearch(id, keyword);
}

/************************************************************************
 * INPUT/OUTPUT AND USER INTERFACE FUNCTIONS
 *
 ************************************************************************/

/*
 * Implement reading strings and symbols
 *
 */

/*
char *clipsRead(void)
{
  char *s = wxGetTextFromUser("Enter text");
  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}
*/

#ifdef USE_CLIPS
char *clipsReadString(void)
{
  char *s = wxGetTextFromUser("Enter text");
  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}
#endif

/*
 * Pass optional message and optional default string
 *
 */

#ifdef USE_CLIPS
char *clipsGetTextFromUser(void)
{
  char *message = "Enter a string";
  char *default_string = "";
  Bool centre = TRUE;

  if (ArgCountCheck("clipsGetTextFromUser", AT_LEAST, 1) != -1)
    message = RtnLexeme(1);

  if (!message)
    return (char *)AddSymbol("");

  int no_args = RtnArgCount();

  if (no_args > 1)
    default_string = RtnLexeme(2);
  if (!default_string)
    return (char *)AddSymbol("");

  if (no_args > 2)
    centre = (Bool)RtnLong(3);

  long parentId = 0;
  if (no_args > 3)
    parentId = RtnLong(4);

  char *s = wxcGetTextFromUser(message, "Input text", default_string, centre, parentId);
  if (s)
    return (char *)AddSymbol(s);
  else
    return (char *)AddSymbol("");
}
#endif

/*
 * Convert string to number
 *
 */

#ifdef USE_CLIPS
double clipsStringToFloat(void)
{
  if (ArgCountCheck("clipsStringToFloat", EXACTLY, 1) == -1)
  {
    ExtensionError << "Error in string-to-float: one string argument required.\n";
    return 0.0;
  }

  char *str = RtnLexeme(1);
  return wxcStringToFloat(str);
}

long clipsStringToLong(void)
{
  if (ArgCountCheck("clipsStringToLong", EXACTLY, 1) == -1)
  {
    ExtensionError << "Error in string-to-long: one string argument required.\n";
    return 0;
  }

  char *str = RtnLexeme(1);
  return wxcStringToLong(str);
}

/*
 * Convert number to string
 *
 */

char *clipsFloatToString(void)
{
  if (ArgCountCheck("clipsFloatToString", EXACTLY, 1) == -1)
  {
    ExtensionError << "Error in float-to-string: one floating point argument required.\n";
    return (char *)AddSymbol("");
  }
  return (char *)AddSymbol(wxcFloatToString(RtnDouble(1)));
}

char *clipsLongToString(void)
{
  if (ArgCountCheck("clipsLongToString", EXACTLY, 1) == -1)
  {
    ExtensionError << "Error in long-to-string: one integer argument required.\n";
    return (char *)AddSymbol("");
  }
  return (char *)AddSymbol(wxcFloatToString(RtnLong(1)));
}

/*
 * Convert string to symbol
 *
 */

char *clipsStringToSymbol(void)
{
  if (ArgCountCheck("clipsStringToSymbol", EXACTLY, 1) == -1)
  {
    ExtensionError << "Error in string-to-symbol: one string argument required.\n";
    return (char *)AddSymbol("");
  }

  char *str = RtnLexeme(1);
  if (!str)
    return (char *)AddSymbol("");
  return (char *)AddSymbol(str);
}

/*
 * Convert symbol to string
 *
 */

char *clipsSymbolToString(void)
{
  if (ArgCountCheck("clipsSymbolToString", EXACTLY, 1) == -1)
  {
    ExtensionError << "Error in symbol-to-string: one symbol argument required.\n";
    return (char *)AddSymbol("");
  }

  char *str = RtnLexeme(1);
  if (!str)
    return (char *)AddSymbol("");
  return (char *)AddSymbol(str);
}
#endif

/*
 * Convert a list of strings or symbols to a sorted list of strings
 *
 */

#ifdef USE_CLIPS
VOID clipsStringSort(DATA_OBJECT_PTR returnValuePtr)
{
  DATA_OBJECT arg;

  if (ArgCountCheck("clipsStringSort", AT_LEAST, 1) == -1)
  {
    ExtensionError << "Error in string-sort: 1st argument must be a multifield of strings or symbols.\n";
    SetMultifieldErrorValue(returnValuePtr);
    return;
  }

  RtnUnknown(1, &arg);

  if (ClipsGetType(arg) != MULTIFIELD)
  {
    ExtensionError << "Error in string-sort: 1st argument must be a multifield of strings or symbols.\n";
    SetMultifieldErrorValue(returnValuePtr);
    return;
  }

  int start = GetpDOBegin(&arg);
  int end = GetpDOEnd(&arg);
  int no_strings = end-start + 1;

  VOID *multifieldPtr = ClipsGetValue(arg);
  wxStringList stringList;

  int i;
  for (i = start; i <= end; i++)
  {
    if ((GetMFType(multifieldPtr, i) == STRING) || (GetMFType(multifieldPtr, i) == STRING))
      stringList.Add(ValueToString(GetMFValue(multifieldPtr, i)));
    else
    {
      ExtensionError << "Error in string-sort: multifield element " << i-start+1 << " must be a string or symbol.\n";
      SetMultifieldErrorValue(returnValuePtr);
      return;
    }
  }
  stringList.Sort();

  VOID *newMultifield = CreateMultifield(no_strings);
  for (i = start; i <= end; i++)
  {
    wxNode *node = stringList.Nth(i-start);
    if (node)
    {
      char *s = (char *)node->Data();

      SetMFType(newMultifield, i, STRING);
      SetMFValue(newMultifield, i, AddSymbol(s));
    }
  }
  SetpType(returnValuePtr, MULTIFIELD);
  SetpValue(returnValuePtr, newMultifield);
  SetpDOBegin(returnValuePtr,start);
  SetpDOEnd(returnValuePtr,end);
  return;
}
#endif

/*
 * Ask user to choose from a menu of strings (message followed by a
 * variable number of string arguments)
 *
 */

#ifndef USE_CLIPS
char *wxcGetChoice(char *message, ClipsExpr& choices, long centre, long parentId)
{
  if (choices.Type() != ClipsList)
    return NULL;

  int n = 0;
  char **choiceStrings = new char *[choices.Number()];
  ClipsExpr *expr = choices.GetFirst();
  while (expr)
  {
    if (expr->Type() == ClipsString || expr->Type() == ClipsWord)
    {
      choiceStrings[n] = (char *)expr->StringValue();
      n ++;
    }
    else if (expr->Type() == ClipsStringClass)
    {
      choiceStrings[n] = (char *)expr->StringClassValue().GetData();
      n ++;
    }
    expr = expr->GetNext();
  }
  wxWindow *parent = NULL;
  if (parentId > 0)
  {
    parent = (wxWindow *)wxGetTypedObject(parentId, wxTYPE_WINDOW);
    if (!parent)
    {
      ExtensionError << "Error in get-choice: could not find parent window " << parentId << ".\n";
      delete[] choiceStrings;
      return NULL;
    }
  }

  char *s = wxGetSingleChoice(message, "Menu", n, choiceStrings, parent, -1, -1, (Bool)centre);
  delete[] choiceStrings;
  return s;
}
#endif

#ifdef USE_CLIPS
char *clipsGetChoice(void)
{
  DATA_OBJECT arg;
  Bool centre = TRUE;
  long parentId = 0;

  if (ArgCountCheck("clipsGetChoice", AT_LEAST, 2) == -1)
    return (char *)AddSymbol("");

  int no_args = RtnArgCount();
  if (no_args > 2)
    centre = (Bool)RtnLong(3);
  if (no_args > 3)
    parentId = RtnLong(4);

  RtnUnknown(2, &arg);

  if (ClipsGetType(arg) != MULTIFIELD)
  {
    ExtensionError << "Error in get-choice: 2nd argument must be a multifield of strings.\n";
    return (char *)AddSymbol("");
  }

  char *message = RtnLexeme(1);
  if (!message)
    return (char *)AddSymbol("");

  int start = GetpDOBegin(&arg);
  int end = GetpDOEnd(&arg);
  int no_strings = end-start + 1;

  VOID *multifieldPtr = ClipsGetValue(arg);

  char **choices = new char *[no_strings];

  for (int i = start; i <= end; i++)
  {
    if ((GetMFType(multifieldPtr, i) == STRING) || (GetMFType(multifieldPtr, i) == SYMBOL))
      choices[i-start] = ValueToString(GetMFValue(multifieldPtr, i));
    else
    {
      ExtensionError << "Error in get-choice: 2nd argument, multifield element " << i-start+1 << " must be a string or symbol.\n";
      delete choices;
      return (char *)AddSymbol("");
    }
  }

  wxWindow *parent = NULL;
  if (parentId > 0)
  {
    parent = (wxWindow *)wxGetTypedObject(parentId, wxTYPE_WINDOW);
    if (!parent)
    {
      ExtensionError << "Error in get-choice: could not find parent window " << parentId << ".\n";
      return NULL;
    }
  }

  char *s = wxGetSingleChoice(message, "Menu", no_strings, choices, parent, -1, -1, centre);
  delete choices;

  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}

char *clipsMessageBox(void)
{
  int no_args = RtnArgCount();
  if (ArgCountCheck("clipsMessageBox", AT_LEAST, 1) == -1)
    return (char *)AddSymbol("");

  char *message = RtnLexeme(1);
  if (!message)
    return (char *)AddSymbol("");

  char *flags = "OK";
  Bool centre = TRUE;
  long parentId = 0;
  char *title = "Message";

  if (no_args > 1)
    flags = RtnLexeme(2);
  if (!flags)
    return (char *)AddSymbol("");

  if (no_args > 2)
    centre = (Bool)RtnLong(3);
  if (no_args > 3)
    parentId = RtnLong(4);
  if (no_args > 4)
    title = RtnLexeme(5);

  char *s = wxcMessageBox(message, flags, centre, parentId, title);
  if (!s)
    return (char *)AddSymbol("");
  else
    return (char *)AddSymbol(s);
}

char *clipsFileSelector(void)
{
  char *message = "Select a file";
  char *def_path = NULL;
  char *def_file = NULL;
  char *def_ext = NULL;
  char *def_wildcard = "*.*";
  long parentId = 0;

  int no_args = RtnArgCount();
  if (no_args > 0)
    message = RtnLexeme(1);
  if (!message)
    return (char *)AddSymbol("");

  if (no_args > 1)
    def_path = RtnLexeme(2);
  if (no_args > 2)
    def_file = RtnLexeme(3);
  if (no_args > 3)
    def_ext = RtnLexeme(4);
  if (no_args > 4)
    def_wildcard = RtnLexeme(5);
  if (!def_wildcard)
    return (char *)AddSymbol("");
  if (no_args > 5)
    parentId = RtnLong(6);

  char *s = wxcFileSelector(message, def_path, def_file, def_ext, def_wildcard, 0, parentId);

  if (s)
    return (char *)AddSymbol(s);
  return (char *)AddSymbol("");
}

/*
 * Yield to windowing system (if appropriate -- i.e. Windows only)
 *
 */

long clipsYield(void)
{
  return wxcYield();
}

/*
 * Sleep for nSecs seconds
 *
 */

long clipsSleep(void)
{
  if (ArgCountCheck("clipsSleep", EXACTLY, 1) == -1)
    return 0;
  return wxcSleep(RtnLong(1));
}

/*
 * Execute a MCI string (Windows only)
 */

char *clipsMCISendString()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsMCISendString", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");
  char *cmd = RtnLexeme(1);
  char *err = wxcMCISendString(cmd);
  if (!err)
    return (char *)AddSymbol("");
  return (char *)AddSymbol(err);
#else
  ExtensionError << "mci-send-string available under Windows only." << "\n";
  return (char *)AddSymbol("");
#endif
}

/*
 * Windows-specific functions
 */
 
Bool clipsHWNDFind()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsHWNDFind", EXACTLY, 1) == -1)
    return FALSE;
  char *title = RtnLexeme(1);
  if (!title)
    return FALSE;
  return (Bool)wxcHWNDFind(title);
#else
  ExtensionError << "hwnd-find available under Windows only." << "\n";
  return FALSE;
#endif
}

Bool clipsHWNDIconize()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsHWNDIconize", AT_LEAST, 1) == -1)
    return FALSE;
  long hWnd = RtnLong(1);
  Bool iconize = TRUE;
  if (RtnArgCount() > 1)
    iconize = (Bool)RtnLong(2);
  return wxcHWNDIconize(hWnd, iconize);
#else
  ExtensionError << "hwnd-iconize available under Windows only." << "\n";
  return FALSE;
#endif
}

Bool clipsHWNDSendMessage()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsHWNDSendMessage", EXACTLY, 4) == -1)
    return FALSE;
  long hWnd = RtnLong(1);
  long msg = RtnLong(2);
  long wParam = RtnLong(3);
  long lParam = RtnLong(4);
  return wxcHWNDSendMessage(hWnd, msg, wParam, lParam);
#else
  ExtensionError << "hwnd-send-message available under Windows only." << "\n";
  return FALSE;
#endif
}

Bool clipsHWNDShow()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsHWNDShow", AT_LEAST, 1) == -1)
    return FALSE;
  long hWnd = RtnLong(1);
  Bool show = TRUE;
  if (RtnArgCount() > 1)
    show = (Bool)RtnLong(2);
  return wxcHWNDShow(hWnd, show);
#else
  ExtensionError << "hwnd-show available under Windows only." << "\n";
  return FALSE;
#endif
}

Bool clipsHWNDRefresh()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsHWNDRefresh", AT_LEAST, 1) == -1)
    return FALSE;
  long hWnd = RtnLong(1);
  Bool eraseBackground = TRUE;
  if (RtnArgCount() > 1)
    eraseBackground = (Bool)RtnLong(2);
  return wxcHWNDRefresh(hWnd, eraseBackground);
#else
  ExtensionError << "hwnd-refresh available under Windows only." << "\n";
  return FALSE;
#endif
}

Bool clipsHWNDMove()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsHWNDMove", AT_LEAST, 5) == -1)
    return FALSE;
  long hWnd = RtnLong(1);
  int x = (int)RtnLong(2);
  int y = (int)RtnLong(3);
  int w = (int)RtnLong(4);
  int h = (int)RtnLong(5);
  Bool repaint = TRUE;
  if (RtnArgCount() > 5)
    repaint = (Bool)RtnLong(6);
  return wxcHWNDMove(hWnd, x, y, w, h, repaint);
#else
  ExtensionError << "hwnd-move available under Windows only." << "\n";
  return FALSE;
#endif
}

Bool clipsHWNDQuit()
{
#ifdef wx_msw
  if (ArgCountCheck("clipsHWNDQuit", EXACTLY, 1) == -1)
    return FALSE;
  long hWnd = RtnLong(1);
  return wxcHWNDQuit(hWnd);
#else
  ExtensionError << "hwnd-quit available under Windows only." << "\n";
  return FALSE;
#endif
}

void clipsBell()
{
  if (ArgCountCheck("clipsBell", EXACTLY, 0) == -1)
    return ;
  wxcBell();
}

long clipsChdir()
{
  if (ArgCountCheck("clipsChdir", EXACTLY, 1) == -1)
    return 0;
  char *dir = RtnLexeme(1);
  if (dir)
    return (long)wxcChdir(dir);
  else
    return 0;
}

long clipsFileExists()
{
  if (ArgCountCheck("clipsFileExists", EXACTLY, 1) == -1)
    return 0;
  char *f = RtnLexeme(1);
  if (f)
    return (long)wxcFileExists(f);
  else
    return 0;
}

long clipsDirExists()
{
  if (ArgCountCheck("clipsDirExists", EXACTLY, 1) == -1)
    return 0;
  char *dir = RtnLexeme(1);
  if (dir)
    return (long)wxcDirExists(dir);
  else
    return 0;
}

/*
 * Write a resource (Windows WIN.INI or Xdefaults)
 *
 */

long clipsWriteResource(void)
{
  if (ArgCountCheck("clipsWriteResource", AT_LEAST, 3) == -1)
    return 0;
  char *section = RtnLexeme(1);
  char *entry = RtnLexeme(2);
  char *value = RtnLexeme(3);
  char *filename = NULL;
  if (RtnArgCount() > 3)
    filename = RtnLexeme(4);
  return wxcWriteResource(section, entry, value, filename);
}

char *clipsGetResource(void)
{
  if (ArgCountCheck("clipsGetResource", AT_LEAST, 2) == -1)
    return (char *)AddSymbol("");
  char *section = RtnLexeme(1);
  char *entry = RtnLexeme(2);
  char *filename = NULL;
  if (RtnArgCount() > 2)
    filename = RtnLexeme(3);
  char *val = wxcGetResource(section, entry, filename);
  if (val)
  {
    char *s = (char *)AddSymbol(val);
    delete[] val;
    return s;
  }
  else
    return (char *)AddSymbol("");
}

/*
 * Write a debug message
 *
 */

long clipsDebugMsg(void)
{
  if (ArgCountCheck("clipsDebugMsg", EXACTLY, 1) == -1)
    return 0;
  char *msg = RtnLexeme(1);
  return wxcDebugMsg(msg);
}

/* Have to implement new batch command since CLIPS assumes using
 * the main loop
 */

void wxExecuteClipsFile(char *filename)
{
  if (FileExists(filename))
  {
    ExecuteBatchFile(filename);
  }
#ifdef wx_x
  else cerr << "Cannot find CLIPS file " << filename << ".\n";
#endif
#ifdef wx_msw
  else
  {
    char buf[400];
    sprintf(buf, "Cannot find CLIPS file %s", filename);
    wxMessageBox(buf, "wxCLIPS problem", wxOK);
  }
#endif
}

void clipsBatch(void)
{
  if (ArgCountCheck("clipsBatch", EXACTLY, 1) == -1)
    return;
  char *filename = RtnLexeme(1);
  if (!filename)
    return;
  wxExecuteClipsFile(filename);
}

/*
 * Begin a 'busy' section
 *
 */

void clipsBeginBusyCursor(void)
{
  if (ArgCountCheck("clipsBeginBusyCursor", EXACTLY, 0) == -1)
    return;
  wxcBeginBusyCursor();
}

/*
 * End a 'busy' section
 *
 */

void clipsEndBusyCursor(void)
{
  if (ArgCountCheck("clipsEndBusyCursor", EXACTLY, 0) == -1)
    return;
  wxcEndBusyCursor();
}

/*
 * System function
 */

long clipsExecute(void)
{
  if (ArgCountCheck("clipsExecute", AT_LEAST, 1) == -1)
    return 0;
  char *s = RtnLexeme(1);
  if (!s)
    return 0;
  Bool sync = FALSE;
  if (RtnArgCount() > 1)
    sync = (Bool)RtnLong(2);
  return wxExecute(s, sync);
}

/*
 * .wxr functions
 */
 
long clipsLoadResourceFile(void)
{
  if (ArgCountCheck("clipsLoadResourceFile", EXACTLY, 1) == -1)
    return 0;
  char *s = RtnLexeme(1);
  if (!s)
    return 0;
  return (long)wxcLoadResourceFile(s);
}

long clipsClearResources(void)
{
  if (ArgCountCheck("clipsClearResources", EXACTLY, 0) == -1)
    return 0;
  wxcClearResources();
  return 1;
}


/*
 * UTILITIES FOR MAPPING IDS <-> COOL INSTANCES
 */

static wxHashTable wxCoolInstanceTable;

long clipsInstanceTableAddEntry(void)
{
  if (ArgCountCheck("clipsInstanceTableAddEntry", EXACTLY, 2) == -1)
    return 0;
  long id = RtnLong(1);
  char *inst = RtnLexeme(2);
  if (!inst)
    return 0;
  wxCoolInstanceTable.Put(id, (wxObject *)copystring(inst));
  return 1;
}

char *clipsInstanceTableGetInstance(void)
{
  if (ArgCountCheck("clipsInstanceTableGetInstance", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");
  long id = RtnLong(1);
  char *s = (char *)wxCoolInstanceTable.Get(id);
  if (!s)
    return (char *)AddSymbol("");
  else
    return (char *)AddSymbol(s);
}

long clipsInstanceTableDeleteEntry(void)
{
  if (ArgCountCheck("clipsInstanceTableDeleteEntry", EXACTLY, 1) == -1)
    return 0;
  long id = RtnLong(1);
  char *s = (char *)wxCoolInstanceTable.Delete(id);
  if (s)
  {
    delete[] s;
    return 1;
  }
  else
    return 0;
}

long clipsWXCLIPSObjectExists(void)
{
  if (ArgCountCheck("clipsWXCLIPSObjectExists", EXACTLY, 1) == -1)
    return 0;
  wxObject *obj = (wxObject *)wxGetTypedObject(RtnLong(1), wxTYPE_OBJECT);
  if (obj)
    return 1;
  else
    return 0;
}

/*
 * Dummy 'declare' for C++ type specification.
 */
 
void clipsDeclare(void)
{
}
#endif

void wxInitClips(void)
{
  // Register language-specific callbacks: in general->specific type
  // order.

  // Any object
  wxRegisterExtensionCallback(wxTYPE_OBJECT, "OnDelete",
     (wxExtensionFunction)wxClipsObjectOnDelete);

  // Timer
  wxRegisterExtensionCallback(wxTYPE_TIMER, "Notify",
     (wxExtensionFunction)wxClipsTimerClassNotify);

  // Windows
  wxRegisterExtensionCallback(wxTYPE_WINDOW, "OnSize",
     (wxExtensionFunction)wxClipsWindowOnSize);
  wxRegisterExtensionCallback(wxTYPE_WINDOW, "OnClose",
     (wxExtensionFunction)wxClipsWindowOnClose);
  wxRegisterExtensionCallback(wxTYPE_WINDOW, "OnChar",
     (wxExtensionFunction)wxClipsWindowOnChar);
  wxRegisterExtensionCallback(wxTYPE_WINDOW, "OnEvent",
     (wxExtensionFunction)wxClipsWindowOnEvent);
  wxRegisterExtensionCallback(wxTYPE_WINDOW, "OnPaint",
     (wxExtensionFunction)wxClipsWindowOnPaint);
  wxRegisterExtensionCallback(wxTYPE_FRAME, "OnMenuCommand",
     (wxExtensionFunction)wxClipsFrameOnMenuCommand);
  wxRegisterExtensionCallback(wxTYPE_FRAME, "OnMenuSelect",
     (wxExtensionFunction)wxClipsFrameOnMenuSelect);
  wxRegisterExtensionCallback(wxTYPE_FRAME, "OnActivate",
     (wxExtensionFunction)wxClipsFrameOnActivate);

  // Toolbar
  wxRegisterExtensionCallback(wxTYPE_TOOLBAR, "OnLeftClick",
     (wxExtensionFunction)wxClipsToolBarClassOnLeftClick);
  wxRegisterExtensionCallback(wxTYPE_TOOLBAR, "OnRightClick",
     (wxExtensionFunction)wxClipsToolBarClassOnRightClick);
  wxRegisterExtensionCallback(wxTYPE_TOOLBAR, "OnMouseEnter",
     (wxExtensionFunction)wxClipsToolBarClassOnMouseEnter);

  // Items
  wxRegisterExtensionCallback(wxTYPE_ITEM, "OnCommand",
     (wxExtensionFunction)wxClipsItemCallback);

  // DDE
  wxRegisterExtensionCallback(wxTYPE_DDE_CONNECTION, "OnAdvise",
     (wxExtensionFunction)wxClipsConnectionOnAdvise);
  wxRegisterExtensionCallback(wxTYPE_DDE_CONNECTION, "OnExecute",
     (wxExtensionFunction)wxClipsConnectionOnExecute);
  wxRegisterExtensionCallback(wxTYPE_DDE_CONNECTION, "OnPoke",
     (wxExtensionFunction)wxClipsConnectionOnPoke);
  wxRegisterExtensionCallback(wxTYPE_DDE_CONNECTION, "OnRequest",
     (wxExtensionFunction)wxClipsConnectionOnRequest);
  wxRegisterExtensionCallback(wxTYPE_DDE_CONNECTION, "OnStartAdvise",
     (wxExtensionFunction)wxClipsConnectionOnStartAdvise);
  wxRegisterExtensionCallback(wxTYPE_DDE_CONNECTION, "OnStopAdvise",
     (wxExtensionFunction)wxClipsConnectionOnStopAdvise);
  wxRegisterExtensionCallback(wxTYPE_DDE_SERVER, "OnAcceptConnection",
     (wxExtensionFunction)wxClipsServerOnAcceptConnection);
  
  // wxApp
#ifdef wx_msw
  wxRegisterExtensionCallback(wxTYPE_APP, "OnCharHook",
     (wxExtensionFunction)wxClipsAppClassOnCharHook);
  wxRegisterExtensionCallback(wxTYPE_WINDOW, "OnCharHook",
     (wxExtensionFunction)wxClipsWindowOnCharHook);
#endif

#ifdef USE_CLIPS
  InitializeCLIPS();
#endif
}

void wxCleanupClips(void)
{
  RouteCommand("(halt)");
}

// Route without echoing return result
void wxRouteNoEcho(char *buf)
{
    Bool oldEcho = clipsEchoOn;
    clipsEchoOn = FALSE;
    RouteCommand(buf);
    SetHaltExecution(CLIPS_FALSE);
    SetEvaluationError(CLIPS_FALSE);
    clipsEchoOn = oldEcho;
}

/*************************************************************/
/* UserFunctions:  The function which informs CLIPS of any   */
/*   user defined functions.  In the default case, there are */
/*   no user defined functions.  To define functions, either */
/*   this function must be replaced by a function with the   */
/*   same name within this file, or this function can be     */
/*   deleted from this file and included in another file.    */
/*   User defined functions may be included in this file or  */
/*   other files.                                            */
/*   Example of redefined UserFunctions:                     */
/*     UserFunctions()                                       */
/*       {                                                   */
/*        DefineFunction("fun1",'i',fun1,"fun1");            */
/*        DefineFunction("other",'f',other,"other");         */
/*       }                                                   */
/*************************************************************/

#ifdef USE_CLIPS
void wxUserFunctions1()
{  
  // Miscellaneous functions
  DefineFunction("batch", 'v', PTIF clipsBatch, "clipsBatch");
  DefineFunction("file-selector", 's', PTIF clipsFileSelector, "clipsFileSelector");
  DefineFunction("float-to-string", 's', PTIF clipsFloatToString, "clipsFloatToString");
  DefineFunction("long-to-string", 's', PTIF clipsLongToString, "clipsLongToString");
  DefineFunction("get-choice", 's', PTIF clipsGetChoice, "clipsGetChoice");
  DefineFunction("get-text-from-user", 's', PTIF clipsGetTextFromUser, "clipsGetTextFromUser");
  DefineFunction("get-platform", 's', PTIF clipsGetPlatform, "clipsGetPlatform");
  DefineFunction("get-os-version", 's', PTIF clipsGetOsVersion, "clipsGetOsVersion");
  DefineFunction("message-box", 'w', PTIF clipsMessageBox, "clipsMessageBox");
  DefineFunction("read-string", 's', PTIF clipsReadString, "clipsReadString");
  DefineFunction("string-to-float", 'd', PTIF clipsStringToFloat, "clipsStringToFloat");
  DefineFunction("string-to-long", 'l', PTIF clipsStringToLong, "clipsStringToLong");
  DefineFunction("string-to-symbol", 'w', PTIF clipsStringToSymbol, "clipsStringToSymbol");
  DefineFunction("symbol-to-string", 's', PTIF clipsSymbolToString, "clipsSymbolToString");
  DefineFunction("yield", 'l', PTIF clipsYield, "clipsYield");
  DefineFunction("debug-msg", 'l', PTIF clipsDebugMsg, "clipsDebugMsg");
  DefineFunction("sleep", 'l', PTIF clipsSleep, "clipsSleep");
  DefineFunction("execute", 'l', PTIF clipsExecute, "clipsExecute");

  DefineFunction("clean-windows", 'v', PTIF clipsCleanWindows, "clipsCleanWindows");

}

void wxUserFunctions2()
{  
  DefineFunction("app-create", 'l', PTIF clipsAppCreate, "clipsAppCreate");
  DefineFunction("app-set-show-frame-on-init", 'v', PTIF clipsAppSetShowFrameOnInit, "clipsAppSetShowFrameOnInit");
  DefineFunction("app-get-show-frame-on-init", 'l', PTIF clipsAppGetShowFrameOnInit, "clipsAppGetShowFrameOnInit");

  DefineFunction("event-get-event-type", 's', PTIF clipsEventGetEventType, "clipsEventGetEventType");

  DefineFunction("command-event-is-selection", 'l', PTIF clipsCommandEventIsSelection, "clipsCommandEventIsSelection");
  DefineFunction("command-event-get-selection", 'l', PTIF clipsCommandEventGetSelection, "clipsCommandEventGetSelection");

  DefineFunction("key-event-get-key-code", 's', PTIF clipsKeyEventGetKeyCode, "clipsKeyEventGetKeyCode");
  DefineFunction("key-event-get-position-x", 'd', PTIF clipsKeyEventPositionX, "clipsKeyEventPositionX");
  DefineFunction("key-event-get-position-y", 'd', PTIF clipsKeyEventPositionY, "clipsKeyEventPositionY");
  DefineFunction("key-event-control-down", 'l', PTIF clipsKeyEventControlDown, "clipsKeyEventControlDown");
  DefineFunction("key-event-shift-down", 'l', PTIF clipsKeyEventShiftDown, "clipsKeyEventShiftDown");
  DefineFunction("key-event-alt-down", 'l', PTIF clipsKeyEventAltDown, "clipsKeyEventAltDown");

  DefineFunction("mouse-event-is-button", 'l', PTIF clipsMouseEventIsButton, "clipsMouseEventIsButton");
  DefineFunction("mouse-event-button-down", 'l', PTIF clipsMouseEventButtonDown, "clipsMouseEventButtonDown");
  DefineFunction("mouse-event-control-down", 'l', PTIF clipsMouseEventControlDown, "clipsMouseEventControlDown");
  DefineFunction("mouse-event-shift-down", 'l', PTIF clipsMouseEventShiftDown, "clipsMouseEventShiftDown");
  DefineFunction("mouse-event-button", 'l', PTIF clipsMouseEventButton, "clipsMouseEventButton");
  DefineFunction("mouse-event-left-down", 'l', PTIF clipsMouseEventLeftDown, "clipsMouseEventLeftDown");
  DefineFunction("mouse-event-middle-down", 'l', PTIF clipsMouseEventMiddleDown, "clipsMouseEventMiddleDown");
  DefineFunction("mouse-event-right-down", 'l', PTIF clipsMouseEventRightDown, "clipsMouseEventRightDown");
  DefineFunction("mouse-event-left-up", 'l', PTIF clipsMouseEventLeftUp, "clipsMouseEventLeftUp");
  DefineFunction("mouse-event-middle-up", 'l', PTIF clipsMouseEventMiddleUp, "clipsMouseEventMiddleUp");
  DefineFunction("mouse-event-right-up", 'l', PTIF clipsMouseEventRightUp, "clipsMouseEventRightUp");
  DefineFunction("mouse-event-dragging", 'l', PTIF clipsMouseEventDragging, "clipsMouseEventDragging");
  DefineFunction("mouse-event-position-x", 'd', PTIF clipsMouseEventPositionX, "clipsMouseEventPositionX");
  DefineFunction("mouse-event-position-y", 'd', PTIF clipsMouseEventPositionY, "clipsMouseEventPositionY");

  DefineFunction("return-result", 'v', PTIF clipsReturnResult, "clipsReturnResult");

  DefineFunction("string-sort", 'm', PTIF clipsStringSort, "clipsStringSort");

  // Timer
  DefineFunction("timer-create", 'l', PTIF clipsTimerCreate, "clipsTimerCreate");
  DefineFunction("timer-start", 'l', PTIF clipsTimerStart, "clipsTimerStart");
  DefineFunction("timer-stop", 'l', PTIF clipsTimerStop, "clipsTimerStop");
  DefineFunction("timer-delete", 'l', PTIF clipsTimerDelete, "clipsTimerDelete");

  // Stopwatch and time
  DefineFunction("start-timer", 'v', PTIF clipsStartTimer, "clipsStartTimer");
  DefineFunction("get-elapsed-time", 'l', PTIF clipsGetElapsedTime, "clipsGetElapsedTime");
  DefineFunction("now", 's', PTIF clipsNow, "clipsNow");

  // Help
  DefineFunction("help-create", 'l', PTIF clipsHelpCreate, "clipsHelpCreate");
  DefineFunction("help-delete", 'l', PTIF clipsHelpDelete, "clipsHelpDelete");
  DefineFunction("help-load-file", 'l', PTIF clipsHelpLoadFile, "clipsHelpLoadFile");
  DefineFunction("help-display-contents", 'l', PTIF clipsHelpDisplayContents, "clipsHelpDisplayContents");
  DefineFunction("help-display-section", 'l', PTIF clipsHelpDisplaySection, "clipsHelpDisplaySection");
  DefineFunction("help-display-block", 'l', PTIF clipsHelpDisplayBlock, "clipsHelpDisplayBlock");
  DefineFunction("help-keyword-search", 'l', PTIF clipsHelpKeywordSearch, "clipsHelpKeywordSearch");

  DefineFunction("mci-send-string", 's', PTIF clipsMCISendString, "clipsMCISendString");
  DefineFunction("bell", 'v', PTIF clipsBell, "clipsBell");
  DefineFunction("chdir", 'l', PTIF clipsChdir, "clipsChdir");
  DefineFunction("file-exists", 'l', PTIF clipsFileExists, "clipsFileExists");
  DefineFunction("dir-exists", 'l', PTIF clipsDirExists, "clipsDirExists");

  DefineFunction("hwnd-find", 'l', PTIF clipsHWNDFind, "clipsHWNDFind");
  DefineFunction("hwnd-show", 'l', PTIF clipsHWNDShow, "clipsHWNDShow");
  DefineFunction("hwnd-iconize", 'l', PTIF clipsHWNDIconize, "clipsHWNDIconize");
  DefineFunction("hwnd-move", 'l', PTIF clipsHWNDMove, "clipsHWNDMove");
  DefineFunction("hwnd-refresh", 'l', PTIF clipsHWNDRefresh, "clipsHWNDRefresh");
  DefineFunction("hwnd-quit", 'l', PTIF clipsHWNDQuit, "clipsHWNDQuit");
  DefineFunction("hwnd-send-message", 'l', PTIF clipsHWNDSendMessage, "clipsHWNDSendMessage");
  
  // Resources
  DefineFunction("write-resource", 'l', PTIF clipsWriteResource, "clipsWriteResource");
  DefineFunction("get-resource", 's', PTIF clipsGetResource, "clipsGetResource");
  
  // .wxr resource system
  DefineFunction("clear-resources", 'l', PTIF clipsClearResources, "clipsClearResources");
  DefineFunction("load-resource-file", 'l', PTIF clipsLoadResourceFile, "clipsLoadResourceFile");

  DefineFunction("begin-busy-cursor", 'v', PTIF clipsBeginBusyCursor, "clipsBeginBusyCursor");
  DefineFunction("end-busy-cursor", 'v', PTIF clipsEndBusyCursor, "clipsEndBusyCursor");
   
  /*
   * C++ compatibility functions
   *
   */
   
  DefineFunction("declare", 'v', PTIF clipsDeclare, "clipsDeclare");
  DefineFunction("wxclips-object-exists", 'l', PTIF clipsWXCLIPSObjectExists, "clipsWXCLIPSObjectExists");

  /*
   * Instance table functions (helper functions for COOL version of wxCLIPS)
   */
  DefineFunction("instance-table-add-entry", 'l', PTIF clipsInstanceTableAddEntry, "clipsInstanceTableAddEntry");
  DefineFunction("instance-table-delete-entry", 'l', PTIF clipsInstanceTableDeleteEntry, "clipsInstanceTableDeleteEntry");
  DefineFunction("instance-table-get-instance", 'o', PTIF clipsInstanceTableGetInstance, "clipsInstanceTableGetInstance");

/*
  DefineFunction("help-load-file", 'l', PTIF clipsHelpLoadFile, "clipsHelpLoadFile");
  DefineFunction("help-display-contents", 'l', PTIF clipsHelpDisplayContents, "clipsHelpDisplayContents");
  DefineFunction("help-display-section", 'l', PTIF clipsHelpDisplaySection, "clipsHelpDisplaySection");
  DefineFunction("help-display-block", 'l', PTIF clipsHelpDisplayBlock, "clipsHelpDisplayBlock");
  DefineFunction("help-keyword-search", 'l', PTIF clipsHelpKeywordSearch, "clipsHelpKeywordSearch");
*/
}

extern void wxGDIUserFunctions();
extern void wxItemUserFunctions();
extern void wxWinUserFunctions();
extern void wxDCUserFunctions();
extern void wxToolbarUserFunctions();
extern void wxDDEUserFunctions();
extern void wxDateUserFunctions();

#if defined(wx_msw) && USE_ODBC && WXODBC
extern void wxODBCUserFunctions();
#endif

void wxUserFunctions()
{
  wxGDIUserFunctions();
  wxDDEUserFunctions();
  wxDCUserFunctions();
  wxItemUserFunctions();
  wxWinUserFunctions();
  wxToolbarUserFunctions();
  wxUserFunctions1();
  wxUserFunctions2();
#if defined(wx_msw) && USE_ODBC && WXODBC
  wxODBCUserFunctions();
#endif
  wxDateUserFunctions();
}
#endif
