/*
 * wxcwin.cc
 * wxWindows command processing: windows
 *
 */

// For compilers that support precompilation, includes "wx.h".
#include "wx_prec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx.h"
#endif

#include "wx_hash.h"
#include "wx_types.h"

#include <iostream.h>
#include <ctype.h>
#include <stdlib.h>

#include "wx_cmds.h"
#include "wxcmdsp.h"

#ifdef USE_CLIPS
#ifndef __WATCOMC__
#include "clips.h"
#endif
#endif

void wxClipsWindowOnSize(Bool *handled, WXID id, int w, int h)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnSize");
  if (func)
  {
    *handled = TRUE;
        
#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld %d %d)", func, id, w, h);

    wxRouteNoEcho(buf);
#else
    (* (void (*)(long, long, long)) func) (id, w, h);
#endif
  }
}

void wxClipsWindowOnChar(Bool *handled, WXID id, long keyCode, long eventId)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnChar");
  if (func)
  {
    *handled = TRUE;

#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld %ld %ld)", func, id, keyCode, eventId);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long, long, long)) func) (id, keyCode, eventId);
#endif
  }
}

void wxClipsWindowOnEvent(Bool *handled, WXID id, long eventId)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnEvent");
  if (func)
  {
    *handled = TRUE;
#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld %ld)", func, id, eventId);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long, long)) func) (id, eventId);
#endif        
  }
}

void wxClipsWindowOnPaint(Bool *handled, WXID id)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnPaint");
  if (func)
  {
    *handled = TRUE;

#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld)", func, id);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long)) func) (id);
#endif
  }
}

#ifdef wx_msw
Bool wxClipsWindowOnCharHook(Bool *handled, WXID id, WXID eventId)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnCharHook");
  if (func)
  {
    *handled = TRUE;

#ifdef USE_CLIPS
    char buf[100];
    clipsReturnType = clipsUNKNOWN;
    sprintf(buf, "(return-result (%s %ld %ld))", func, id, eventId);
    wxRouteNoEcho(buf);
#else
    clipsReturnLong = (* (long (*)(long)) func) (id, eventId);
#endif
    if (clipsReturnType != clipsLONG)
    {
      ExtensionError << "OnCharHook callback should return an integer (1 or 0)!\n";
      return FALSE;
    }
    else
    {
      *handled = TRUE;
      return (Bool)clipsReturnLong;
    }
  }
  return FALSE;
}
#endif

void wxClipsFrameOnMenuCommand(Bool *handled, WXID id, int commandId)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnMenuCommand");
  if (func)
  {
    *handled = TRUE;
#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld %d)", func, id, commandId);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long, int)) func) (id, commandId);
#endif
  }
}

void wxClipsFrameOnMenuSelect(Bool *handled, WXID id, int commandId)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnMenuSelect");
  if (func)
  {
    *handled = TRUE;
#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld %d)", func, id, commandId);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long, int)) func) (id, commandId);
#endif
  }
}

void wxClipsFrameOnActivate(Bool *handled, WXID id, Bool active)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnActivate");
  if (func)
  {
    *handled = TRUE;
#ifdef USE_CLIPS
    char buf[100];
    sprintf(buf, "(%s %ld %d)", func, id, active);
    wxRouteNoEcho(buf);
#else
    (* (void (*)(long, int)) func) (id, active);
#endif
  }
}

Bool wxClipsWindowOnClose(Bool *handled, WXID id)
{
  wxClipsFunction func = wxClipsFindCallback(id, "OnClose");
  if (func)
  {
    *handled = TRUE;
#ifdef USE_CLIPS
    char buf[200];
    clipsReturnType = clipsUNKNOWN;
    sprintf(buf, "(return-result (%s %ld))", func, id);
    wxRouteNoEcho(buf);
#else
    clipsReturnLong = (* (long (*)(long)) func) (id);
    clipsReturnType = clipsLONG;
#endif        
    if (clipsReturnType != clipsLONG)
    {
      ExtensionError << "OnClose callback should return an integer (1 or 0)!\n";
    }
    else return (Bool)clipsReturnLong;
  }
  *handled = TRUE;
  return TRUE;
}

/************************************************************************
 * GENERIC WINDOW FUNCTIONS
 *
 ************************************************************************/

// Add a callback to an object
Bool wxcObjectAddCallback(WXID id, char *eventType, wxClipsFunction callback)
{
  if (!eventType || !callback)
    return FALSE;

  wxObject *obj = (wxObject *)wxGetTypedObject(id, wxTYPE_OBJECT);
  if (obj)
  {
    wxClipsAddCallback(id, eventType, callback);
    return TRUE;
  }
  else
  {
    ExtensionError << "Error in object-add-callback: could not find object " << id << ".\n";
    return FALSE;
  }
}

// Add a callback to a window
#ifdef USE_CLIPS
long clipsObjectAddCallback()
{
  if (ArgCountCheck("wxcObjectAddCallback", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);

  char *event_type = RtnLexeme(2);
  char *callback = RtnLexeme(3);

  return (long)wxcObjectAddCallback(id, event_type, callback);
}

long clipsWindowDelete()
{
  if (ArgCountCheck("clipsWindowDelete", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowDelete(id);
}

long clipsWindowShow()
{
  if (ArgCountCheck("clipsWindowShow", AT_LEAST, 1) == -1)
    return 0;

  long id = RtnLong(1);
  Bool show = TRUE;
  if (RtnArgCount() > 1)
    show = (Bool)RtnLong(2);
  return (long)wxcWindowShow(id, show);
}

long clipsWindowEnable()
{
  if (ArgCountCheck("clipsWindowEnable", AT_LEAST, 1) == -1)
    return 0;

  long id = RtnLong(1);
  Bool enable = TRUE;
  if (RtnArgCount() > 1)
    enable = (Bool)RtnLong(2);
  return (long)wxcWindowEnable(id, enable);
}

long clipsWindowGetParent()
{
  if (ArgCountCheck("clipsWindowGetParent", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  return wxcWindowGetParent(id);
}

char *clipsWindowGetName()
{
  if (ArgCountCheck("clipsWindowGetName", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

  char *name = wxcWindowGetName(RtnLong(1));
  if (name)
    return (char *)AddSymbol(name);
  else
    return (char *)AddSymbol("");
}

// Get window width
long clipsWindowGetWidth()
{
  if (ArgCountCheck("clipsWindowGetWidth", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowGetWidth(id);
}

long clipsWindowGetHeight()
{
  if (ArgCountCheck("clipsWindowGetHeight", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowGetHeight(id);
}

long clipsWindowGetClientWidth()
{
  if (ArgCountCheck("clipsWindowGetClientWidth", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowGetClientWidth(id);
}

long clipsWindowGetClientHeight()
{
  if (ArgCountCheck("clipsWindowGetClientHeight", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowGetClientHeight(id);
}

long clipsWindowGetX()
{
  if (ArgCountCheck("clipsWindowGetX", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowGetX(id);
}

long clipsWindowGetY()
{
  if (ArgCountCheck("clipsWindowGetY", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowGetY(id);
}

long clipsWindowSetSize()
{
  if (ArgCountCheck("clipsWindowSetSize", EXACTLY, 5) == -1)
    return 0;

  long id = RtnLong(1);
  int x = (int) RtnLong(2);
  int y = (int) RtnLong(3);
  int w = (int) RtnLong(4);
  int h = (int) RtnLong(5);
  return (long)wxcWindowSetSize(id, x, y, w, h);
}

long clipsWindowPopupMenu()
{
  if (ArgCountCheck("clipsWindowPopupMenu", EXACTLY, 4) == -1)
    return 0;

  long id = RtnLong(1);
  long menuId = RtnLong(2);
  float x = (float) RtnDouble(3);
  float y = (float) RtnDouble(4);
  return (long)wxcWindowPopupMenu(id, menuId, x, y);
}

long clipsWindowSetClientSize()
{
  if (ArgCountCheck("clipsWindowSetClientSize", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  int w = (int) RtnLong(2);
  int h = (int) RtnLong(3);

  return (long)wxcWindowSetClientSize(id, w, h);
}

long clipsWindowSetFocus()
{
  if (ArgCountCheck("clipsWindowSetFocus", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowSetFocus(id);
}

long clipsWindowSetCursor()
{
  if (ArgCountCheck("clipsWindowSetCursor", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long cursorId = RtnLong(2);

  return (long)wxcWindowSetCursor(id, cursorId);
}

long clipsWindowFit()
{
  if (ArgCountCheck("clipsWindowFit", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcWindowFit(id);
}

long clipsWindowCentre()
{
  if (ArgCountCheck("clipsWindowCentre", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();
  long id = RtnLong(1);
  int direction = wxHORIZONTAL;
  char *stringDir = "wxBOTH";
  if (no_args > 1)
    stringDir = RtnLexeme(2);

  return (long)wxcWindowCentre(id, stringDir);
}

long clipsWindowMakeModal()
{
  if (ArgCountCheck("clipsWindowMakeModal", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long modal = RtnLong(2);

  return (long)wxcWindowMakeModal(id, (Bool)modal);
}

/************************************************************************
 * DIALOG AND PANEL ITEM FUNCTIONS
 *
 ************************************************************************/

long clipsDialogBoxCreate()
{
  if (ArgCountCheck("clipsDialogBoxCreate", AT_LEAST, 2) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *title = RtnLexeme(2);

  if (!title)
    return -1;

  // Set up defaults
  Bool modal = TRUE;
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  char *style = "wxSYSTEM_MENU | wxCAPTION | wxTHICK_FRAME";
  char *name = "dialogBox";

  if (no_args > 2)
    modal = (Bool)RtnLong(3);

  if (no_args > 3)
    x = (int) RtnLong(4);

  if (no_args > 4)
    y = (int) RtnLong(5);

  if (no_args > 5)
    width = (int) RtnLong(6);

  if (no_args > 6)
    height = (int) RtnLong(7);

  if (no_args > 7)
    style = RtnLexeme(8);

  if (no_args > 8)
    name = RtnLexeme(9);

  return wxcDialogBoxCreate(parent_id, title, modal, x, y, width, height, style, name);
}

long clipsDialogBoxCreateFromResource()
{
  if (ArgCountCheck("clipsDialogBoxCreateFromResource", EXACTLY, 2) == -1)
    return 0;

  long parentId = RtnLong(1);
  char *resName = RtnLexeme(2);

  return wxcDialogBoxCreateFromResource(parentId, resName);
}

long clipsDialogBoxIsModal()
{
  if (ArgCountCheck("clipsDialogBoxIsModal", EXACTLY, 1) == -1)
    return 0;
  return (long)wxcDialogBoxIsModal(RtnLong(1));
}

long clipsDialogBoxSetModal()
{
  if (ArgCountCheck("clipsDialogBoxSetModal", EXACTLY, 2) == -1)
    return 0;
  wxcDialogBoxSetModal(RtnLong(1), (Bool)RtnLong(2));
  return 1;
}


long clipsPanelCreate()
{
  if (ArgCountCheck("clipsPanelCreate", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  char *style = "";
  char *name = "panel";

  if (no_args > 1)
    x = (int) RtnLong(2);

  if (no_args > 2)
    y = (int) RtnLong(3);

  if (no_args > 3)
    width = (int) RtnLong(4);

  if (no_args > 4)
    height = (int) RtnLong(5);

  if (no_args > 5)
    style = RtnLexeme(6);

  if (no_args > 6)
    name = RtnLexeme(7);

  return wxcPanelCreate(parent_id, x, y, width, height, style, name);
}

long clipsPanelCreateFromResource()
{
  if (ArgCountCheck("clipsPanelCreateFromResource", EXACTLY, 2) == -1)
    return 0;

  long parentId = RtnLong(1);
  char *resName = RtnLexeme(2);

  return wxcPanelCreateFromResource(parentId, resName);
}

long clipsPanelSetLabelPosition()
{
  if (ArgCountCheck("clipsPanelSetLabelPosition", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *position = RtnLexeme(2);
  if (!position)
    return 0;
  return (long)wxcPanelSetLabelPosition(id, position);
}

long clipsPanelSetLabelFont()
{
  if (ArgCountCheck("clipsPanelSetLabelFont", EXACTLY, 2) == -1)
    return 0;

  return (long)wxcPanelSetLabelFont(RtnLong(1), RtnLong(2));
}

long clipsPanelSetButtonFont()
{
  if (ArgCountCheck("clipsPanelSetButtonFont", EXACTLY, 2) == -1)
    return 0;

  return (long)wxcPanelSetButtonFont(RtnLong(1), RtnLong(2));
}

long clipsPanelNewLine()
{
  if (ArgCountCheck("clipsPanelNewLine", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcPanelNewLine(id);
}

long clipsCanvasCreate()
{
  if (ArgCountCheck("clipsCanvasCreate", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  char *style = "wxRETAINED";
  char *name = "canvas";

  if (no_args > 1)
    x = (int) RtnLong(2);

  if (no_args > 2)
    y = (int) RtnLong(3);

  if (no_args > 3)
    width = (int) RtnLong(4);

  if (no_args > 4)
    height = (int) RtnLong(5);

  if (no_args > 5)
    style = RtnLexeme(6);

  if (no_args > 6)
    name = RtnLexeme(7);

  return wxcCanvasCreate(parent_id, x, y, width, height, style, name);
}

long clipsCanvasSetScrollbars()
{
  if (ArgCountCheck("clipsCanvasSetScrollbars", EXACTLY, 7) == -1)
    return 0;

  long id = RtnLong(1);
  long horiz = RtnLong(2);
  long vert = RtnLong(3);
  long x_length = RtnLong(4);
  long y_length = RtnLong(5);
  long x_page = RtnLong(6);
  long y_page = RtnLong(7);
  return (long)wxcCanvasSetScrollbars(id, (int)horiz, (int)vert, (int)x_length, (int)y_length,
     (int)x_page, (int)y_page);
}

long clipsCanvasScroll()
{
  if (ArgCountCheck("clipsCanvasScroll", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  long x_pos = RtnLong(2);
  long y_pos = RtnLong(3);

  return (long)wxcCanvasScroll(id, (int)x_pos, (int)y_pos);
}

long clipsCanvasGetDC()
{
  if (ArgCountCheck("clipsCanvasGetDC", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);

  return wxcCanvasGetDC(id);
}

long clipsCanvasOnChar()
{
  if (ArgCountCheck("clipsCanvasOnChar", EXACTLY, 2) == -1)
    return 0;

  wxcCanvasOnChar(RtnLong(1), RtnLong(2));
  return 1;
}

long clipsTextWindowCreate()
{
  if (ArgCountCheck("clipsTextWindowCreate", AT_LEAST, 1) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);

  // Set up defaults
  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;
  char *style = "wxNATIVE_IMPL";
  char *name = "textWindow";

  if (no_args > 1)
    x = (int) RtnLong(2);

  if (no_args > 2)
    y = (int) RtnLong(3);

  if (no_args > 3)
    width = (int) RtnLong(4);

  if (no_args > 4)
    height = (int) RtnLong(5);

  if (no_args > 5)
    style = RtnLexeme(6);

  if (no_args > 6)
    name = RtnLexeme(7);

  return wxcTextWindowCreate(parent_id, x, y, width, height, style, name);
}

long clipsTextWindowLoadFile()
{
  if (ArgCountCheck("clipsTextWindowLoadFile", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *filename = RtnLexeme(2);
  return (long)wxcTextWindowLoadFile(id, filename);
}

long clipsTextWindowSaveFile()
{
  if (ArgCountCheck("clipsTextWindowSaveFile", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *filename = RtnLexeme(2);
  return (long)wxcTextWindowSaveFile(id, filename);
}

VOID clipsTextWindowWrite()
{
  if (ArgCountCheck("clipsTextWindowWrite", AT_LEAST, 1) == -1)
    return;

  long id = RtnLong(1);

  DATA_OBJECT temp;

  for (int i = 2; i <= RtnArgCount(); i++)
  {
    RtnUnknown(i, &temp);

    switch (ClipsGetType(temp))
    {
      case INTEGER:
      {
        wxcTextWindowWrite(id, (long)DOToLong(temp));
        break;
      }
      case STRING:
      case SYMBOL:
      {
        wxcTextWindowWrite(id, DOToString(temp));
        break;
      }
      case FLOAT:
      {
        wxcTextWindowWrite(id, DOToDouble(temp));
        break;
      }
      default:
      {
        ExtensionError << "Error in text-window-write: unrecognized type.\n";
        break;
      }
    }
  }
}

long clipsTextWindowModified()
{
  if (ArgCountCheck("clipsTextWindowModifed", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcTextWindowModified(id);
}

long clipsTextWindowClear()
{
  if (ArgCountCheck("clipsTextWindowClear", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcTextWindowClear(id);
}

VOID clipsTextWindowDiscardEdits()
{
  if (ArgCountCheck("clipsTextWindowDiscardEdits", EXACTLY, 1) == -1)
    return;

  long id = RtnLong(1);
  wxcTextWindowDiscardEdits(id);
}

long clipsTextWindowCopy()
{
  if (ArgCountCheck("clipsTextWindowCopy", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcTextWindowCopy(id);
}

long clipsTextWindowPaste()
{
  if (ArgCountCheck("clipsTextWindowPaste", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcTextWindowPaste(id);
}

long clipsTextWindowCut()
{
  if (ArgCountCheck("clipsTextWindowCut", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcTextWindowCut(id);
}

char *clipsTextWindowGetContents()
{
  if (ArgCountCheck("clipsTextWindowGetContents", EXACTLY, 1) == -1)
    return (char *)AddSymbol("");

  long id = RtnLong(1);
  char *s = wxcTextWindowGetContents(id);
  if (s)
    return (char *)AddSymbol(s);
  else
    return (char *)AddSymbol("");
}

long clipsTextWindowSetEditable()
{
  if (ArgCountCheck("clipsTextWindowSetEditable", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcTextWindowSetEditable(id, (Bool)RtnLong(2));
}

long clipsTextWindowOnChar()
{
  if (ArgCountCheck("clipsTextWindowOnChar", EXACTLY, 2) == -1)
    return 0;

  wxcTextWindowOnChar(RtnLong(1), RtnLong(2));
  return 1;
}

/*
 * New JACS Nov. 1995
 */
 
long clipsTextWindowReplace()
{
  if (ArgCountCheck("clipsTextWindowReplace", EXACTLY, 4) == -1)
    return 0;

  long id = RtnLong(1);
  long from = RtnLong(2);
  long to = RtnLong(3);
  char *text = RtnLexeme(4);
  return (long)wxcTextWindowReplace(id, from, to, text);
}

long clipsTextWindowRemove()
{
  if (ArgCountCheck("clipsTextWindowRemove", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  long from = RtnLong(2);
  long to = RtnLong(3);
  return (long)wxcTextWindowRemove(id, from, to);
}

long clipsTextWindowSetSelection()
{
  if (ArgCountCheck("clipsTextWindowSetSelection", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  long from = RtnLong(2);
  long to = RtnLong(3);
  return (long)wxcTextWindowSetSelection(id, from, to);
}

long clipsTextWindowGetLastPosition()
{
  if (ArgCountCheck("clipsTextWindowGetLastPosition", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return (long)wxcTextWindowGetLastPosition(id);
}

long clipsTextWindowShowPosition()
{
  if (ArgCountCheck("clipsTextWindowShowPosition", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long pos = RtnLong(2);
  return (long)wxcTextWindowShowPosition(id, pos);
}

long clipsTextWindowPositionToLine()
{
  if (ArgCountCheck("clipsTextWindowPositionToLine", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long pos = RtnLong(2);
  return (long)wxcTextWindowPositionToLine(id, pos);
}

long clipsTextWindowPositionToChar()
{
  if (ArgCountCheck("clipsTextWindowPositionToChar", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long pos = RtnLong(2);
  return (long)wxcTextWindowPositionToChar(id, pos);
}

long clipsTextWindowXYToPosition()
{
  if (ArgCountCheck("clipsTextWindowXYToPosition", EXACTLY, 3) == -1)
    return 0;

  long id = RtnLong(1);
  long x = RtnLong(2);
  long y = RtnLong(3);
  return (long)wxcTextWindowXYToPosition(id, x, y);
}

long clipsTextWindowSetInsertionPoint()
{
  if (ArgCountCheck("clipsTextWindowSetInsertionPoint", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long pos = RtnLong(2);
  wxcTextWindowSetInsertionPoint(id, pos);
  return 1;
}

long clipsTextWindowGetInsertionPoint()
{
  if (ArgCountCheck("clipsTextWindowGetInsertionPoint", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcTextWindowGetInsertionPoint(id);
}

long clipsTextWindowGetLineLength()
{
  if (ArgCountCheck("clipsTextWindowGetLineLength", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long line = RtnLong(2);
  return wxcTextWindowGetLineLength(id, line);
}

char *clipsTextWindowGetLineText()
{
  if (ArgCountCheck("clipsTextWindowGetLineText", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long line = RtnLong(2);
  char *s =  wxcTextWindowGetLineText(id, line);
  if (!s)
    return (char *)AddSymbol("");
  else
    return (char *)AddSymbol(s);
}

long clipsTextWindowGetNumberOfLines()
{
  if (ArgCountCheck("clipsTextWindowGetNumberOfLines", EXACTLY, 1) == -1)
    return 0;

  long id = RtnLong(1);
  return wxcTextWindowGetNumberOfLines(id);
}

/************************************************************************
 * FRAME FUNCTIONS
 *
 ************************************************************************/

long clipsFrameCreate()
{
  if (ArgCountCheck("clipsFrameCreate", AT_LEAST, 2) == -1)
    return 0;

  int no_args = RtnArgCount();

  long parent_id = RtnLong(1);
  char *title = RtnLexeme(2);

  // Set up defaults

  int x = -1;
  int y = -1;
  int width = -1;
  int height = -1;

  if (no_args > 2)
    x = (int) RtnLong(3);

  if (no_args > 3)
    y = (int) RtnLong(4);

  if (no_args > 4)
    width = (int) RtnLong(5);

  if (no_args > 5)
    height = (int) RtnLong(6);

  char *style = "wxSDI | wxDEFAULT_FRAME";
  char *name = "frame";
  
  if (no_args > 6)
    style = RtnLexeme(7);
  if (no_args > 7)
    name = RtnLexeme(8);
  return wxcFrameCreate(parent_id, title, x, y, width, height, style, name);
}

long clipsFrameSetMenuBar()
{
  if (ArgCountCheck("clipsFrameSetMenuBar", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long menubar_id = RtnLong(2);
  return (long)wxcFrameSetMenuBar(id, menubar_id);
}

long clipsFrameSetToolBar()
{
  if (ArgCountCheck("clipsFrameSetToolBar", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long toolbar_id = RtnLong(2);
  return (long)wxcFrameSetToolBar(id, toolbar_id);
}

long clipsFrameCreateStatusLine()
{
  if (ArgCountCheck("clipsFrameCreateStatusLine", AT_LEAST, 1) == -1)
    return 0;

  long id = RtnLong(1);
  long n = 1;
  if (RtnArgCount() > 1)
    n = RtnLong(2);
  return (long)wxcFrameCreateStatusLine(id, n);
}

long clipsFrameSetStatusText()
{
  if (ArgCountCheck("clipsFrameSetStatusText", AT_LEAST, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *text = RtnLexeme(2);
  int i = 0;
  if (RtnArgCount() > 2)
    i = RtnLong(3);
  return (long)wxcFrameSetStatusText(id, text, i);
}

long clipsFrameSetTitle()
{
  if (ArgCountCheck("clipsFrameSetTitle", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  char *text = RtnLexeme(2);
  return (long)wxcFrameSetTitle(id, text);
}

long clipsFrameSetIcon()
{
  if (ArgCountCheck("clipsFrameSetIcon", EXACTLY, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long iconId = RtnLong(2);
  return (long)wxcFrameSetIcon(id, iconId);
}

long clipsFrameIconize()
{
  if (ArgCountCheck("clipsFrameIconize", AT_LEAST, 2) == -1)
    return 0;

  long id = RtnLong(1);
  long iconize = TRUE;

  if (RtnArgCount() > 1)
    iconize = RtnLong(2);
  return (long)wxcFrameIconize(id, (Bool)iconize);
}

void clipsFrameOnSize()
{
  if (ArgCountCheck("clipsFrameOnSize", EXACTLY, 3) == -1)
    return;

  wxcFrameOnSize(RtnLong(1), (int)RtnLong(2), (int)RtnLong(3));
}

#endif


/*************************************************************/
/* UserFunctions:  The function which informs CLIPS of any   */
/*   user defined functions.  In the default case, there are */
/*   no user defined functions.  To define functions, either */
/*   this function must be replaced by a function with the   */
/*   same name within this file, or this function can be     */
/*   deleted from this file and included in another file.    */
/*   User defined functions may be included in this file or  */
/*   other files.                                            */
/*   Example of redefined UserFunctions:                     */
/*     UserFunctions()                                       */
/*       {                                                   */
/*        DefineFunction("fun1",'i',fun1,"fun1");            */
/*        DefineFunction("other",'f',other,"other");         */
/*       }                                                   */
/*************************************************************/

#ifdef USE_CLIPS
void wxWinUserFunctions()
{  
  DefineFunction("canvas-create", 'l', PTIF clipsCanvasCreate, "clipsCanvasCreate");
  DefineFunction("canvas-set-scrollbars", 'l', PTIF clipsCanvasSetScrollbars, "clipsCanvasSetScrollbars");
  DefineFunction("canvas-scroll", 'l', PTIF clipsCanvasScroll, "clipsCanvasScroll");
  DefineFunction("canvas-get-dc", 'l', PTIF clipsCanvasGetDC, "clipsCanvasGetDC");
  DefineFunction("canvas-on-char", 'l', PTIF clipsCanvasOnChar, "clipsCanvasOnChar");

  DefineFunction("dialog-box-create", 'l', PTIF clipsDialogBoxCreate, "clipsDialogBoxCreate");
  DefineFunction("dialog-box-create-from-resource", 'l', PTIF clipsDialogBoxCreateFromResource, "clipsDialogBoxCreateFromResource");
  DefineFunction("dialog-box-is-modal", 'l', PTIF clipsDialogBoxIsModal, "clipsDialogBoxIsModal");
  DefineFunction("dialog-box-set-modal", 'l', PTIF clipsDialogBoxSetModal, "clipsDialogBoxSetModal");

  DefineFunction("frame-create", 'l', PTIF clipsFrameCreate, "clipsFrameCreate");
  DefineFunction("frame-create-status-line", 'l', PTIF clipsFrameCreateStatusLine, "clipsFrameCreateStatusLine");
  DefineFunction("frame-iconize", 'l', PTIF clipsFrameIconize, "clipsFrameIconize");
  DefineFunction("frame-set-icon", 'l', PTIF clipsFrameSetIcon, "clipsFrameSetIcon");
  DefineFunction("frame-set-menu-bar", 'l', PTIF clipsFrameSetMenuBar, "clipsFrameSetMenuBar");
  DefineFunction("frame-set-tool-bar", 'l', PTIF clipsFrameSetToolBar, "clipsFrameSetToolBar");
  DefineFunction("frame-set-status-text", 'l', PTIF clipsFrameSetStatusText, "clipsFrameSetStatusText");
  DefineFunction("frame-set-title", 'l', PTIF clipsFrameSetTitle, "clipsFrameSetTitle");
  DefineFunction("frame-on-size", 'v', PTIF clipsFrameOnSize, "clipsFrameOnSize");

  DefineFunction("panel-create", 'l', PTIF clipsPanelCreate, "clipsPanelCreate");
  DefineFunction("panel-create-from-resource", 'l', PTIF clipsPanelCreateFromResource, "clipsPanelCreateFromResource");
  DefineFunction("panel-set-label-position", 'l', PTIF clipsPanelSetLabelPosition, "clipsPanelSetLabelPosition");
  DefineFunction("panel-set-label-font", 'l', PTIF clipsPanelSetLabelFont, "clipsPanelSetLabelFont");
  DefineFunction("panel-set-button-font", 'l', PTIF clipsPanelSetButtonFont, "clipsPanelSetButtonFont");
  DefineFunction("panel-new-line", 'l', PTIF clipsPanelNewLine, "clipsPanelNewLine");

  DefineFunction("text-window-clear", 'l', PTIF clipsTextWindowClear, "clipsTextWindowClear");
  DefineFunction("text-window-create", 'l', PTIF clipsTextWindowCreate, "clipsTextWindowCreate");
  DefineFunction("text-window-load-file", 'l', PTIF clipsTextWindowLoadFile, "clipsTextWindowLoadFile");
  DefineFunction("text-window-save-file", 'l', PTIF clipsTextWindowSaveFile, "clipsTextWindowSaveFile");
  DefineFunction("text-window-modified", 'l', PTIF clipsTextWindowModified, "clipsTextWindowModified");
  DefineFunction("text-window-discard-edits", 'v', PTIF clipsTextWindowDiscardEdits, "clipsTextWindowDiscardEdits");
  DefineFunction("text-window-write", 'v', PTIF clipsTextWindowWrite, "clipsTextWindowWrite");
  DefineFunction("text-window-copy", 'l', PTIF clipsTextWindowCopy, "clipsTextWindowCopy");
  DefineFunction("text-window-cut", 'l', PTIF clipsTextWindowCut, "clipsTextWindowCut");
  DefineFunction("text-window-paste", 'l', PTIF clipsTextWindowPaste, "clipsTextWindowPaste");
  DefineFunction("text-window-get-contents", 's', PTIF clipsTextWindowGetContents, "clipsTextWindowGetContents");
  DefineFunction("text-window-set-editable", 'l', PTIF clipsTextWindowSetEditable, "clipsTextWindowSetEditable");
  DefineFunction("text-window-on-char", 'l', PTIF clipsTextWindowOnChar, "clipsTextWindowOnChar");

  // New stuff Nov. 1995 (yes this line'll look silly when Nov. 1995 is ancient history...)
  DefineFunction("text-window-replace", 'l', PTIF clipsTextWindowReplace, "clipsTextWindowReplace");
  DefineFunction("text-window-remove", 'l', PTIF clipsTextWindowRemove, "clipsTextWindowRemove");
  DefineFunction("text-window-set-selection", 'l', PTIF clipsTextWindowSetSelection, "clipsTextWindowSetSelection");
  DefineFunction("text-window-get-last-position", 'l', PTIF clipsTextWindowGetLastPosition, "clipsTextWindowGetLastPosition");
  DefineFunction("text-window-show-position", 'l', PTIF clipsTextWindowShowPosition, "clipsTextWindowShowPosition");
  DefineFunction("text-window-position-to-line", 'l', PTIF clipsTextWindowPositionToLine, "clipsTextWindowPositionToLine");
  DefineFunction("text-window-position-to-char", 'l', PTIF clipsTextWindowPositionToChar, "clipsTextWindowPositionToChar");
  DefineFunction("text-window-set-insertion-point", 'l', PTIF clipsTextWindowSetInsertionPoint, "clipsTextWindowSetInsertionPoint");
  DefineFunction("text-window-get-insertion-point", 'l', PTIF clipsTextWindowGetInsertionPoint, "clipsTextWindowGetInsertionPoint");
  DefineFunction("text-window-get-line-length", 'l', PTIF clipsTextWindowGetLineLength, "clipsTextWindowGetLineLength");
  DefineFunction("text-window-get-line-length", 'l', PTIF clipsTextWindowGetLineLength, "clipsTextWindowGetLineLength");
  DefineFunction("text-window-get-line-text", 's', PTIF clipsTextWindowGetLineText, "clipsTextWindowGetLineText");
  DefineFunction("text-window-get-number-of-lines", 'l', PTIF clipsTextWindowGetNumberOfLines, "clipsTextWindowGetNumberOfLines");
  DefineFunction("text-window-xy-to-position", 'l', PTIF clipsTextWindowXYToPosition, "clipsTextWindowXYToPosition");

  DefineFunction("object-add-callback", 'l', PTIF clipsObjectAddCallback, "clipsObjectAddCallback");
  DefineFunction("window-add-callback", 'l', PTIF clipsObjectAddCallback, "clipsObjectAddCallback");
  DefineFunction("window-delete", 'l', PTIF clipsWindowDelete, "clipsWindowDelete");
  DefineFunction("window-get-parent", 'l', PTIF clipsWindowGetParent, "clipsWindowGetParent");
  DefineFunction("window-show", 'l', PTIF clipsWindowShow, "clipsWindowShow");
  DefineFunction("window-enable", 'l', PTIF clipsWindowEnable, "clipsWindowEnable");

  DefineFunction("window-get-name", 's', PTIF clipsWindowGetName, "clipsWindowGetName");
  DefineFunction("window-get-x", 'l', PTIF clipsWindowGetX, "clipsWindowGetX");
  DefineFunction("window-get-y", 'l', PTIF clipsWindowGetY, "clipsWindowGetY");
  DefineFunction("window-get-width", 'l', PTIF clipsWindowGetWidth, "clipsWindowGetWidth");
  DefineFunction("window-get-height", 'l', PTIF clipsWindowGetHeight, "clipsWindowGetHeight");
  DefineFunction("window-get-client-width", 'l', PTIF clipsWindowGetClientWidth, "clipsWindowGetClientWidth");
  DefineFunction("window-get-client-height", 'l', PTIF clipsWindowGetClientHeight, "clipsWindowGetClientHeight");
  DefineFunction("window-set-size", 'l', PTIF clipsWindowSetSize, "clipsWindowSetSize");
  DefineFunction("window-set-client-size", 'l', PTIF clipsWindowSetClientSize, "clipsWindowSetClientSize");
  DefineFunction("window-set-focus", 'l', PTIF clipsWindowSetFocus, "clipsWindowSetFocus");
  DefineFunction("window-set-cursor", 'l', PTIF clipsWindowSetCursor, "clipsWindowSetCursor");
  DefineFunction("window-centre", 'l', PTIF clipsWindowCentre, "clipsWindowCentre");
  DefineFunction("window-fit", 'l', PTIF clipsWindowFit, "clipsWindowFit");
  DefineFunction("window-make-modal", 'l', PTIF clipsWindowMakeModal, "clipsWindowMakeModal");
  DefineFunction("window-popup-menu", 'l', PTIF clipsWindowPopupMenu, "clipsWindowPopupMenu");
}
#endif

