/*
 * wxctbar.cc
 * wxWindows command processing: toolbar
 *
 */

// For compilers that support precompilation, includes "wx.h".
#include "wx_prec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx.h"
#endif

#include "wx_hash.h"
#include "wx_types.h"

#include "wx_tbar.h"

#ifdef wx_msw
#include "wx_bbar.h"
#endif

#include <iostream.h>
#include <ctype.h>
#include <stdlib.h>

#include "wxextend.h"
#include "extbarp.h"

// extern "C" int clipsEchoOn;
extern ExtensionErrorClass ExtensionError;

void wxClipsToolBarClass::OnPaint (void)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnPaint");
  if (func)
    {
      Bool handled = FALSE;

      (*(void (*)(Bool *, long)) func) (&handled, (long) this);

      if (!handled)
        wxToolBar::OnPaint ();
    }
  else
    wxToolBar::OnPaint ();
}

Bool wxClipsToolBarClass::OnLeftClick (int toolIndex, Bool toggleDown)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnLeftClick");
  if (func)
    {
      Bool handled = FALSE;

      Bool returnValue = (*(Bool (*)(Bool *, long, int, Bool))
		       func) (&handled, (long) this, toolIndex, toggleDown);

      if (!handled)
        return wxToolBar::OnLeftClick (toolIndex, toggleDown);
      else
	return returnValue;
    }
  else
    return wxToolBar::OnLeftClick (toolIndex, toggleDown);
}

void wxClipsToolBarClass::OnRightClick (int toolIndex, float x, float y)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnRightClick");
  if (func)
    {
      Bool handled = FALSE;

      (*(void (*)(Bool *, long, int, float, float)) func)
	(&handled, (long) this, toolIndex, x, y);

      if (!handled)
        wxToolBar::OnRightClick (toolIndex, x, y);
    }
  else
    wxToolBar::OnRightClick (toolIndex, x, y);
}

void wxClipsToolBarClass::OnMouseEnter (int toolIndex)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnMouseEnter");
  if (func)
    {
      Bool handled = FALSE;

      (*(void (*)(Bool *, long, int))
       func) (&handled, (long) this, toolIndex);

      if (!handled)
        wxToolBar::OnMouseEnter (toolIndex);
    }
  else
    wxToolBar::OnMouseEnter (toolIndex);
}

/*
 * Buttonbar implementation (almost identical to toolbar, but
 * the C++ class is different, so we need to repeat the code).
 */

void wxClipsButtonBarClass::OnPaint (void)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnPaint");
  if (func)
    {
      Bool handled = FALSE;

      (*(void (*)(Bool *, long)) func) (&handled, (long) this);

      if (!handled)
#ifdef wx_msw
        wxButtonBar::OnPaint ();
#else
        wxToolBar::OnPaint ();
#endif
    }
  else
#ifdef wx_msw
    wxButtonBar::OnPaint ();
#else
    wxToolBar::OnPaint ();
#endif
}

Bool wxClipsButtonBarClass::OnLeftClick (int toolIndex, Bool toggleDown)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnLeftClick");
  if (func)
    {
      Bool handled = FALSE;

      Bool returnValue = (*(Bool (*)(Bool *, long, int, Bool))
		       func) (&handled, (long) this, toolIndex, toggleDown);

      if (!handled)
#ifdef wx_msw
        return wxButtonBar::OnLeftClick (toolIndex, toggleDown);
#else
        return wxToolBar::OnLeftClick (toolIndex, toggleDown);
#endif
      else
	return returnValue;
    }
  else
#ifdef wx_msw
    return wxButtonBar::OnLeftClick (toolIndex, toggleDown);
#else
    return wxToolBar::OnLeftClick (toolIndex, toggleDown);
#endif
}

void wxClipsButtonBarClass::OnRightClick (int toolIndex, float x, float y)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnRightClick");
  if (func)
    {
      Bool handled = FALSE;

      (*(void (*)(Bool *, long, int, float, float)) func)
	(&handled, (long) this, toolIndex, x, y);

      if (!handled)
#ifdef wx_msw
        wxButtonBar::OnRightClick (toolIndex, x, y);
#else
        wxToolBar::OnRightClick (toolIndex, x, y);
#endif
    }
  else
#ifdef wx_msw
    wxButtonBar::OnRightClick (toolIndex, x, y);
#else
    wxToolBar::OnRightClick (toolIndex, x, y);
#endif
}

void wxClipsButtonBarClass::OnMouseEnter (int toolIndex)
{
  wxExtensionFunction func = wxFindExtensionCallback (wxTYPE_TOOLBAR, "OnMouseEnter");
  if (func)
    {
      Bool handled = FALSE;

      (*(void (*)(Bool *, long, int))
       func) (&handled, (long) this, toolIndex);

      if (!handled)
#ifdef wx_msw
        wxButtonBar::OnMouseEnter (toolIndex);
#else
        wxToolBar::OnMouseEnter (toolIndex);
#endif
    }
  else
#ifdef wx_msw
    wxButtonBar::OnMouseEnter (toolIndex);
#else
    wxToolBar::OnMouseEnter (toolIndex);
#endif
}

// Create a new toolbar
WXID wxcToolbarCreate (WXID parent_id, int x, int y, int w, int h, long style,
		  int orientation, int rowsOrColumns, Bool createButtons, char *name)
{
  wxWindow *parent = NULL;

  if (parent_id != 0)
    {
      parent = (wxWindow *) wxGetTypedObject (parent_id, wxTYPE_WINDOW);
      if (!parent)
	{
	  ExtensionError << "Error in toolbar-create: could not find parent window " << parent_id << ".\n";
	  return 0;
	}
    }

  wxToolBar *toolbar = NULL;
  if (createButtons)
    {
      toolbar = new wxClipsButtonBarClass (parent, x, y, w, h, style,
					   orientation, rowsOrColumns, name);
      wxStoreThingInTable ((long) toolbar, toolbar, wxTYPE_BUTTONBAR);
    }
  else
    {
      toolbar = new wxClipsToolBarClass (parent, x, y, w, h, style,
					 orientation, rowsOrColumns, name);
      wxStoreThingInTable ((long) toolbar, toolbar, wxTYPE_TOOLBAR);
    }

  return (long) toolbar;
}

// Create a new toolbar (string version)
WXID wxcToolbarCreate (WXID parent_id, int x, int y, int w, int h, char *styleString,
		  char *orientation, int rowsOrColumns, Bool createButtons, char *name)
{
  if (!orientation)
    return 0;
  int orient = wxVERTICAL;
  if (strcmp (orientation, "wxVERTICAL") == 0)
    orient = wxVERTICAL;
  else
    orient = wxHORIZONTAL;

  wxWindow *parent = NULL;

  if (parent_id != 0)
    {
      parent = (wxWindow *) wxGetTypedObject (parent_id, wxTYPE_WINDOW);
      if (!parent)
	{
	  ExtensionError << "Error in toolbar-create: could not find parent window " << parent_id << ".\n";
	  return 0;
	}
    }
  long style = wxExtensionParseBitList (styleString);

  wxToolBar *toolbar = NULL;
  if (createButtons)
    {
      toolbar = new wxClipsButtonBarClass (parent, x, y, w, h, style,
					   orient, rowsOrColumns, name);
      wxStoreThingInTable ((long) toolbar, toolbar, wxTYPE_BUTTONBAR);
    }
  else
    {
      toolbar = new wxClipsToolBarClass (parent, x, y, w, h, style,
					 orient, rowsOrColumns, name);
      wxStoreThingInTable ((long) toolbar, toolbar, wxTYPE_TOOLBAR);
    }

  return (long) toolbar;
}

// Add a tool
Bool wxcToolbarAddTool (WXID id, int index, WXID bitmapId1, WXID bitmapId2, Bool toggle,
		   double x, double y, long clientData)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-add-tool: could not find toolbar " << id << ".\n";
      return FALSE;
    }
  wxBitmap *bitmap1 = (wxBitmap *) wxGetTypedObject (bitmapId1, wxTYPE_BITMAP);
  wxBitmap *bitmap2 = NULL;
  if (!bitmap1)
    {
      ExtensionError << "Error in toolbar-add-tool: could not find bitmap " << bitmapId1 << ".\n";
      return FALSE;
    }
  if (bitmapId2 != 0)
    {
      bitmap2 = (wxBitmap *) wxGetTypedObject (bitmapId2, wxTYPE_BITMAP);
      if (!bitmap2)
	{
	  ExtensionError << "Error in toolbar-add-tool: could not find bitmap " << bitmapId2 << ".\n";
	  return FALSE;
	}
    }
  (void) toolbar->AddTool (index, bitmap1, bitmap2, (Bool) toggle, (float) x, (float) y, (wxObject *) clientData);
  return TRUE;
}

Bool wxcToolbarClearTools (WXID id)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-clear-tools: could not find toolbar " << id << ".\n";
      return 0;
    }
  (void) toolbar->ClearTools ();
  return TRUE;
}

Bool wxcToolbarEnableTool (WXID id, int index, Bool enable)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-enable-tool: could not find toolbar " << id << ".\n";
      return 0;
    }
  toolbar->EnableTool (index, enable);
  return TRUE;
}

Bool wxcToolbarToggleTool (WXID id, int index, Bool toggle)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-toggle-tool: could not find toolbar " << id << ".\n";
      return FALSE;
    }
  toolbar->ToggleTool (index, toggle);
  return TRUE;
}

Bool wxcToolbarLayout (WXID id)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-layout: could not find toolbar " << id << ".\n";
      return FALSE;
    }
  (void) toolbar->Layout ();
  return TRUE;
}

double wxcToolbarGetMaxWidth (WXID id)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-get-max-width: could not find toolbar " << id << ".\n";
      return 0.0;
    }
  float x, y;
  toolbar->GetMaxSize (&x, &y);
  return (double) x;
}

double wxcToolbarGetMaxHeight (WXID id)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-get-max-height: could not find toolbar " << id << ".\n";
      return 0.0;
    }
  float x, y;
  toolbar->GetMaxSize (&x, &y);
  return (double) y;
}

Bool wxcToolbarOnPaint (WXID id)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-on-paint: could not find toolbar " << id << ".\n";
      return FALSE;
    }
  toolbar->OnPaint ();
  return TRUE;
}

long wxcToolbarGetToolClientData (WXID id, int index)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-get-tool-client-data: could not find toolbar " << id << ".\n";
      return 0;
    }
  return (long) toolbar->GetToolClientData (index);
}

Bool wxcToolbarGetToolEnabled (WXID id, int index)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-get-enabled: could not find toolbar " << id << ".\n";
      return FALSE;
    }
  return toolbar->GetToolEnabled (index);
}

Bool wxcToolbarGetToolState (WXID id, int index)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-get-state: could not find toolbar " << id << ".\n";
      return FALSE;
    }
  return toolbar->GetToolState (index);
}

Bool wxcToolbarSetMargins (WXID id, double x, double y)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-set-margins: could not find toolbar " << id << ".\n";
      return FALSE;
    }
  toolbar->SetMargins ((float) x, (float) y);
  return TRUE;
}

Bool wxcToolbarSetDefaultSize (WXID id, int x, int y)
{
  wxToolBar *toolbar = (wxToolBar *) wxGetTypedObject (id, wxTYPE_TOOLBAR);
  if (!toolbar)
    {
      ExtensionError << "Error in toolbar-set-default-size: could not find toolbar " << id << ".\n";
      return FALSE;
    }
#ifdef wx_msw
  if (wxSubType (toolbar->__type, wxTYPE_BUTTONBAR))
    ((wxButtonBar *) toolbar)->SetDefaultSize ((float) x, (float) y);
#endif
  return TRUE;
}
